package sun.invoke.util;

import java.lang.reflect.Array;
import java.util.Arrays;

public enum Wrapper
{
    BOOLEAN(Boolean.class, Boolean.TYPE, 'Z', Boolean.valueOf(false), new boolean[0], Wrapper.Format.unsigned(1)), 
    BYTE(Byte.class, Byte.TYPE, 'B', Byte.valueOf((byte)0), new byte[0], Wrapper.Format.signed(8)), 
    SHORT(Short.class, Short.TYPE, 'S', Short.valueOf((short)0), new short[0], Wrapper.Format.signed(16)), 
    CHAR(Character.class, Character.TYPE, 'C', Character.valueOf('\0'), new char[0], Wrapper.Format.unsigned(16)), 
    INT(Integer.class, Integer.TYPE, 'I', Integer.valueOf(0), new int[0], Wrapper.Format.signed(32)), 
    LONG(Long.class, Long.TYPE, 'J', Long.valueOf(0L), new long[0], Wrapper.Format.signed(64)), 
    FLOAT(Float.class, Float.TYPE, 'F', Float.valueOf(0.0f), new float[0], Wrapper.Format.floating(32)), 
    DOUBLE(Double.class, Double.TYPE, 'D', Double.valueOf(0.0), new double[0], Wrapper.Format.floating(64)), 
    OBJECT(Object.class, Object.class, 'L', null, new Object[0], Wrapper.Format.other(1)), 
    VOID(Void.class, Void.TYPE, 'V', null, null, Wrapper.Format.other(0));
    
    private final Class<?> wrapperType;
    private final Class<?> primitiveType;
    private final char basicTypeChar;
    private final Object zero;
    private final Object emptyArray;
    private final int format;
    private final String wrapperSimpleName;
    private final String primitiveSimpleName;
    private static final Wrapper[] FROM_PRIM;
    private static final Wrapper[] FROM_WRAP;
    private static final Wrapper[] FROM_CHAR;
    static final /* synthetic */ boolean $assertionsDisabled;
    
    private Wrapper(Class<?> wrapperType, Class<?> primitiveType, char basicTypeChar, Object zero, Object emptyArray, int format) {
        this.wrapperType = wrapperType;
        this.primitiveType = primitiveType;
        this.basicTypeChar = basicTypeChar;
        this.zero = zero;
        this.emptyArray = emptyArray;
        this.format = format;
        this.wrapperSimpleName = wrapperType.getSimpleName();
        this.primitiveSimpleName = primitiveType.getSimpleName();
    }
    
    public String detailString() {
        return this.wrapperSimpleName + Arrays.asList(new Object[] { this.wrapperType, this.primitiveType, Character.valueOf(this.basicTypeChar), this.zero, "0x" + Integer.toHexString(this.format) });
    }
    
    public int bitWidth() {
        return this.format >> 2 & 0x3FF;
    }
    
    public int stackSlots() {
        return this.format >> 0 & 0x3;
    }
    
    public boolean isSingleWord() {
        return (this.format & 0x1) != 0x0;
    }
    
    public boolean isDoubleWord() {
        return (this.format & 0x2) != 0x0;
    }
    
    public boolean isNumeric() {
        return (this.format & 0xFFFFFFFC) != 0x0;
    }
    
    public boolean isIntegral() {
        return this.isNumeric() && this.format < 4225;
    }
    
    public boolean isSubwordOrInt() {
        return this.isIntegral() && this.isSingleWord();
    }
    
    public boolean isSigned() {
        return this.format < 0;
    }
    
    public boolean isUnsigned() {
        return this.format >= 5 && this.format < 4225;
    }
    
    public boolean isFloating() {
        return this.format >= 4225;
    }
    
    public boolean isOther() {
        return (this.format & 0xFFFFFFFC) == 0x0;
    }
    
    public boolean isConvertibleFrom(Wrapper o) {
        if (this == o) {
            return true;
        }
        if (this.compareTo(o) < 0) {
            return false;
        }
        if ((this.format & o.format & 0xFFFFF000) == 0x0) {
            return this.isOther() || o.format == 65;
        }
        assert !(!this.isSigned());
        assert !(!o.isSigned());
        return true;
    }
    
    private static boolean checkConvertibleFrom() {
        Wrapper[] values = values();
        int length = values.length;
        int i = 0;
        while (i < length) {
            Wrapper wrapper = values[i];
            assert !(!wrapper.isConvertibleFrom(wrapper));
            assert !(!Wrapper.VOID.isConvertibleFrom(wrapper));
            if (wrapper != Wrapper.VOID) {
                assert !(!Wrapper.OBJECT.isConvertibleFrom(wrapper));
                assert !wrapper.isConvertibleFrom(Wrapper.VOID);
            }
            if (wrapper != Wrapper.CHAR) {
                assert !Wrapper.CHAR.isConvertibleFrom(wrapper);
                if (!wrapper.isConvertibleFrom(Wrapper.INT) && !Wrapper.$assertionsDisabled && wrapper.isConvertibleFrom(Wrapper.CHAR)) {
                    throw new AssertionError();
                }
            }
            if (wrapper != Wrapper.BOOLEAN) {
                assert !Wrapper.BOOLEAN.isConvertibleFrom(wrapper);
                if (wrapper != Wrapper.VOID && wrapper != Wrapper.OBJECT && !Wrapper.$assertionsDisabled && wrapper.isConvertibleFrom(Wrapper.BOOLEAN)) {
                    throw new AssertionError();
                }
            }
            if (wrapper.isSigned()) {
                for (Wrapper o : values()) {
                    if (wrapper != o) {
                        if (o.isFloating()) {
                            assert !wrapper.isConvertibleFrom(o);
                        }
                        else if (o.isSigned()) {
                            if (wrapper.compareTo(o) < 0) {
                                assert !wrapper.isConvertibleFrom(o);
                            }
                            else
                                assert !(!wrapper.isConvertibleFrom(o));
                        }
                    }
                }
            }
            if (wrapper.isFloating()) {
                for (Wrapper o2 : values()) {
                    if (wrapper != o2) {
                        if (o2.isSigned()) {
                            assert !(!wrapper.isConvertibleFrom(o2));
                        }
                        else if (o2.isFloating()) {
                            if (wrapper.compareTo(o2) < 0) {
                                assert !wrapper.isConvertibleFrom(o2);
                            }
                            else
                                assert !(!wrapper.isConvertibleFrom(o2));
                        }
                    }
                }
            }
            ++i;
        }
        return true;
    }
    
    public Object zero() {
        return this.zero;
    }
    
    public <T> T zero(Class<T> clazz) {
        return this.convert(this.zero, clazz);
    }
    
    public static Wrapper forPrimitiveType(Class<?> obj) {
        Wrapper primitiveType = findPrimitiveType(obj);
        if (primitiveType != null) {
            return primitiveType;
        }
        if (obj.isPrimitive()) {
            throw new InternalError();
        }
        throw newIllegalArgumentException("not primitive: " + obj);
    }
    
    static Wrapper findPrimitiveType(Class<?> clazz) {
        Wrapper wrapper = Wrapper.FROM_PRIM[hashPrim(clazz)];
        if (wrapper != null && wrapper.primitiveType == clazz) {
            return wrapper;
        }
        return null;
    }
    
    public static Wrapper forWrapperType(Class<?> obj) {
        Wrapper wrapperType = findWrapperType(obj);
        if (wrapperType != null) {
            return wrapperType;
        }
        Wrapper[] values = values();
        int length = values.length;
        int i = 0;
        while (i < length) {
            if (values[i].wrapperType == obj) {
                throw new InternalError();
            }
            ++i;
        }
        throw newIllegalArgumentException("not wrapper: " + obj);
    }
    
    static Wrapper findWrapperType(Class<?> clazz) {
        Wrapper wrapper = Wrapper.FROM_WRAP[hashWrap(clazz)];
        if (wrapper != null && wrapper.wrapperType == clazz) {
            return wrapper;
        }
        return null;
    }
    
    public static Wrapper forBasicType(char c) {
        Wrapper wrapper = Wrapper.FROM_CHAR[hashChar(c)];
        if (wrapper != null && wrapper.basicTypeChar == c) {
            return wrapper;
        }
        Wrapper[] values = values();
        int length = values.length;
        int i = 0;
        while (i < length) {
            Wrapper wrapper2 = values[i];
            if (wrapper.basicTypeChar == c) {
                throw new InternalError();
            }
            ++i;
        }
        throw newIllegalArgumentException("not basic type char: " + c);
    }
    
    public static Wrapper forBasicType(Class<?> clazz) {
        if (clazz.isPrimitive()) {
            return forPrimitiveType(clazz);
        }
        return Wrapper.OBJECT;
    }
    
    private static int hashPrim(Class<?> clazz) {
        String name = clazz.getName();
        if (name.length() < 3) {
            return 0;
        }
        return (name.charAt(0) + name.charAt(2)) % 16;
    }
    
    private static int hashWrap(Class<?> clazz) {
        String name = clazz.getName();
        assert 10 == "java.lang.".length();
        if (name.length() < 13) {
            return 0;
        }
        return (3 * name.charAt(11) + name.charAt(12)) % 16;
    }
    
    private static int hashChar(char c) {
        return (c + (c >> 1)) % 16;
    }
    
    public Class<?> primitiveType() {
        return this.primitiveType;
    }
    
    public Class<?> wrapperType() {
        return this.wrapperType;
    }
    
    public <T> Class<T> wrapperType(Class<T> clazz) {
        if (clazz == this.wrapperType) {
            return clazz;
        }
        if (clazz == this.primitiveType || this.wrapperType == Object.class || clazz.isInterface()) {
            return forceType(this.wrapperType, clazz);
        }
        throw newClassCastException(clazz, this.primitiveType);
    }
    
    private static ClassCastException newClassCastException(Class<?> obj, Class<?> obj2) {
        return new ClassCastException(obj + " is not compatible with " + obj2);
    }
    
    public static <T> Class<T> asWrapperType(Class<T> clazz) {
        if (clazz.isPrimitive()) {
            return forPrimitiveType(clazz).wrapperType(clazz);
        }
        return clazz;
    }
    
    public static <T> Class<T> asPrimitiveType(Class<T> clazz) {
        Wrapper wrapperType = findWrapperType(clazz);
        if (wrapperType != null) {
            return forceType(wrapperType.primitiveType(), clazz);
        }
        return clazz;
    }
    
    public static boolean isWrapperType(Class<?> clazz) {
        return findWrapperType(clazz) != null;
    }
    
    public static boolean isPrimitiveType(Class<?> clazz) {
        return clazz.isPrimitive();
    }
    
    public static char basicTypeChar(Class<?> clazz) {
        if (!clazz.isPrimitive()) {
            return 'L';
        }
        return forPrimitiveType(clazz).basicTypeChar();
    }
    
    public char basicTypeChar() {
        return this.basicTypeChar;
    }
    
    public String wrapperSimpleName() {
        return this.wrapperSimpleName;
    }
    
    public String primitiveSimpleName() {
        return this.primitiveSimpleName;
    }
    
    public <T> T cast(Object o, Class<T> clazz) {
        return this.convert(o, clazz, true);
    }
    
    public <T> T convert(Object o, Class<T> clazz) {
        return this.convert(o, clazz, false);
    }
    
    private <T> T convert(Object o, Class<T> clazz, boolean b) {
        if (this == Wrapper.OBJECT) {
            assert !clazz.isPrimitive();
            if (!clazz.isInterface()) {
                clazz.cast(o);
            }
            return o;
        }
        else {
            Class<T> wrapperType = this.wrapperType(clazz);
            if (wrapperType.isInstance(o)) {
                return wrapperType.cast(o);
            }
            if (!b) {
                Class<Object> class1 = o.getClass();
                Wrapper wrapperType2 = findWrapperType(class1);
                if (wrapperType2 == null || !this.isConvertibleFrom(wrapperType2)) {
                    throw newClassCastException(wrapperType, class1);
                }
            }
            else if (o == null) {
                return this.zero;
            }
            Object wrap = this.wrap(o);
            assert ((wrap == null) ? Void.class : wrap.getClass()) == wrapperType;
            return wrap;
        }
    }
    
    static <T> Class<T> forceType(Class<?> obj, Class<T> obj2) {
        if (obj != obj2 && (!obj.isPrimitive() || forPrimitiveType(obj) != findWrapperType(obj2)) && (!obj2.isPrimitive() || forPrimitiveType(obj2) != findWrapperType(obj)) && (obj != Object.class || obj2.isPrimitive())) {
            System.out.println(obj + " <= " + obj2);
        }
        assert obj == Object.class && !obj2.isPrimitive();
        return obj;
    }
    
    public Object wrap(Object o) {
        switch (this.basicTypeChar) {
            case 'L': {
                return o;
            }
            case 'V': {
                return null;
            }
            default: {
                Number numberValue = numberValue(o);
                switch (this.basicTypeChar) {
                    case 'I': {
                        return Integer.valueOf(numberValue.intValue());
                    }
                    case 'J': {
                        return Long.valueOf(numberValue.longValue());
                    }
                    case 'F': {
                        return Float.valueOf(numberValue.floatValue());
                    }
                    case 'D': {
                        return Double.valueOf(numberValue.doubleValue());
                    }
                    case 'S': {
                        return Short.valueOf((short)numberValue.intValue());
                    }
                    case 'B': {
                        return Byte.valueOf((byte)numberValue.intValue());
                    }
                    case 'C': {
                        return Character.valueOf((char)numberValue.intValue());
                    }
                    case 'Z': {
                        return Boolean.valueOf(boolValue(numberValue.byteValue()));
                    }
                    default: {
                        throw new InternalError("bad wrapper");
                    }
                }
            }
        }
    }
    
    public Object wrap(int n) {
        if (this.basicTypeChar == 'L') {
            return Integer.valueOf(n);
        }
        switch (this.basicTypeChar) {
            case 'L': {
                throw newIllegalArgumentException("cannot wrap to object type");
            }
            case 'V': {
                return null;
            }
            case 'I': {
                return Integer.valueOf(n);
            }
            case 'J': {
                return Long.valueOf((long)n);
            }
            case 'F': {
                return Float.valueOf((float)n);
            }
            case 'D': {
                return Double.valueOf((double)n);
            }
            case 'S': {
                return Short.valueOf((short)n);
            }
            case 'B': {
                return Byte.valueOf((byte)n);
            }
            case 'C': {
                return Character.valueOf((char)n);
            }
            case 'Z': {
                return Boolean.valueOf(boolValue((byte)n));
            }
            default: {
                throw new InternalError("bad wrapper");
            }
        }
    }
    
    private static Number numberValue(Object o) {
        if ((o instanceof Number)) {
            return (Number)o;
        }
        if ((o instanceof Character)) {
            return Integer.valueOf((int)(char)o);
        }
        if ((o instanceof Boolean)) {
            return Integer.valueOf(((boolean)o) ? 1 : 0);
        }
        return (Number)o;
    }
    
    private static boolean boolValue(byte b) {
        return (byte)(b & 0x1) != 0;
    }
    
    private static RuntimeException newIllegalArgumentException(String str, Object obj) {
        return newIllegalArgumentException(str + obj);
    }
    
    private static RuntimeException newIllegalArgumentException(String s) {
        return new IllegalArgumentException(s);
    }
    
    public Object makeArray(int length) {
        return Array.newInstance(this.primitiveType, length);
    }
    
    public Class<?> arrayType() {
        return this.emptyArray.getClass();
    }
    
    public void copyArrayUnboxing(Object[] array, int n, Object obj, int n2, int n3) {
        if (obj.getClass() != this.arrayType()) {
            this.arrayType().cast(obj);
        }
        for (int i = 0; i < n3; ++i) {
            Array.set(obj, i + n2, this.convert(array[i + n], this.primitiveType));
        }
    }
    
    public void copyArrayBoxing(Object obj, int n, Object[] array, int n2, int n3) {
        if (obj.getClass() != this.arrayType()) {
            this.arrayType().cast(obj);
        }
        int i = 0;
        while (i < n3) {
            Object value = Array.get(obj, i + n);
            assert value.getClass() == this.wrapperType;
            array[i + n2] = value;
            ++i;
        }
    }
    
    static {
        assert !(!checkConvertibleFrom());
        FROM_PRIM = new Wrapper[16];
        FROM_WRAP = new Wrapper[16];
        FROM_CHAR = new Wrapper[16];
        Wrapper[] values = values();
        int length = values.length;
        int i = 0;
        while (i < length) {
            Wrapper wrapper = values[i];
            int hashPrim = hashPrim(wrapper.primitiveType);
            int hashWrap = hashWrap(wrapper.wrapperType);
            int hashChar = hashChar(wrapper.basicTypeChar);
            assert Wrapper.FROM_PRIM[hashPrim] == null;
            assert Wrapper.FROM_WRAP[hashWrap] == null;
            assert Wrapper.FROM_CHAR[hashChar] == null;
            Wrapper.FROM_PRIM[hashPrim] = wrapper;
            Wrapper.FROM_WRAP[hashWrap] = wrapper;
            Wrapper.FROM_CHAR[hashChar] = wrapper;
            ++i;
        }
        return;
    }
    
    private abstract static class Format
    {
        static final int SLOT_SHIFT = 0;
        static final int SIZE_SHIFT = 2;
        static final int KIND_SHIFT = 12;
        static final int SIGNED = -4096;
        static final int UNSIGNED = 0;
        static final int FLOATING = 4096;
        static final int SLOT_MASK = 3;
        static final int SIZE_MASK = 1023;
        static final int INT = -3967;
        static final int SHORT = -4031;
        static final int BOOLEAN = 5;
        static final int CHAR = 65;
        static final int FLOAT = 4225;
        static final int VOID = 0;
        static final int NUM_MASK = -4;
        static final /* synthetic */ boolean $assertionsDisabled;
        
        static int format(int n, int n2, int n3) {
            assert n >> 12 << 12 == n;
            assert (n2 & n2 - 1) == 0x0;
            Label_0108: {
                if (!Wrapper.Format.$assertionsDisabled) {
                    if (n == -4096) {
                        if (n2 > 0) {
                            break Label_0108;
                        }
                    }
                    else if (n == 0) {
                        if (n2 > 0) {
                            break Label_0108;
                        }
                    }
                    else if (n == 4096) {
                        if (n2 != 32) {
                            if (n2 == 64) {
                                break Label_0108;
                            }
                        }
                        else {
                            break Label_0108;
                        }
                    }
                    throw new AssertionError();
                }
            }
            assert !((n3 != 2) ? (n3 != 1 || n2 > 32) : (n2 != 64));
            return n | n2 << 2 | n3 << 0;
        }
        
        static int signed(int n) {
            return format(-4096, n, (n > 32) ? 2 : 1);
        }
        
        static int unsigned(int n) {
            return format(0, n, (n > 32) ? 2 : 1);
        }
        
        static int floating(int n) {
            return format(4096, n, (n > 32) ? 2 : 1);
        }
        
        static int other(int n) {
            return n << 0;
        }
    }
}
